# FILE: procs.tcl
#
# A collection of Tcl procedures (not Oc_Classes) which are part of the
# Oc extension
#
# Need to split this into files likely to load together

# Returns the absolute, direct pathname of its argument
proc Oc_DirectPathname { pathname } {
    global Oc_DirectPathnameCache
    set canCdTo [file dirname $pathname]
    set rest [file tail $pathname]
    switch -exact -- $rest {
        .	-
        .. {
            set canCdTo [file join $canCdTo $rest]
            set rest ""
        }
    }
    if {[string match absolute [file pathtype $canCdTo]]} {
        set index $canCdTo
    } else {
        set index [file join [pwd] $canCdTo]
    }
    if {[info exists Oc_DirectPathnameCache($index)]} {
        return [file join $Oc_DirectPathnameCache($index) $rest]
    }
    # A directory must be "executable" for us to be able to 'cd' to it
    while {![file executable $canCdTo]} {
        switch -exact -- [file tail $canCdTo] {
            "" {
                # $canCdTo is the root directory, and we can't cd to it.
                # This means we know the direct pathname, even though we
                # can't cd to it or any of its ancestors.
                set Oc_DirectPathnameCache($index) $canCdTo	;# = '/'
                return [file join $Oc_DirectPathnameCache($index) $rest]
            }
            . {
                # Do nothing.  Leave $rest unchanged
            }
            .. {
                # NOMAC: Assuming '..' means 'parent directory'
                # Don't want to shift '..' onto $rest.
                # Make recursive call instead.
                set Oc_DirectPathnameCache($index) [file dirname \
                        [Oc_DirectPathname [file dirname $canCdTo]]]
                return [file join $Oc_DirectPathnameCache($index) $rest]
            }
            default {
                ;# Shift one path component from $canCdTo to $rest
                set rest [file join [file tail $canCdTo] $rest]
            }
        }
        set canCdTo [file dirname $canCdTo]
        set index [file dirname $index]
    }
    set savedir [pwd]
    if {[catch {cd $canCdTo} msg]} {
        return -code error "Can't determine direct pathname for 
	$pathname:
	$msg"
    }
    set Oc_DirectPathnameCache($index) [pwd]
    cd $savedir
    return [file join $Oc_DirectPathnameCache($index) $rest]
}

proc Oc_MakePortHeader {varinfo outfile} {
    # Eventually make calls to objects representing local configuration.
    puts "Updating [file join [pwd] $outfile] ..."
    global tcl_platform
    # See if we can tell what platform we are on
    set systemtype unknown  ;# For local use
    if {![info exists tcl_platform(platform)]} {
	set systemtype unix  ;# Can't tell, so assume unix
    } else {
	if {[string compare $tcl_platform(platform) "unix"] == 0} {
	    set systemtype unix
	} elseif {[string compare $tcl_platform(platform) "windows"] == 0} {
	    set systemtype windows
	} elseif {[string compare $tcl_platform(platform) "macintosh"] == 0} {
	    set systemtype macintosh
	}
    }

    # Run varinfo and parse output
    set varlist [exec $varinfo]
    foreach vartype {char short int long float double} {
	set varorder($vartype) -1   ;# Safety
	regexp "$vartype is *(\[0-9\]*) bytes wide *Byte order: *(\[0-9\]*)" \
		$varlist tempmatch varsize($vartype) varorder($vartype)
    }
    set varsize(pointer) -1
    regexp {void \* is *([0-9]*)} $varlist tempmatch varsize(pointer)

    foreach varwidth {FLT DBL} {
        regexp "\n${varwidth}_EPSILON: (\[^\n\]+)" \
                $varlist tempmatch vareps($varwidth)
        regexp "\nSQRT_${varwidth}_EPSILON: (\[^\n\]+)" \
                $varlist tempmatch vareps(SQRT_$varwidth)
    }

    # Open output file
    if { [string compare $outfile stdout] == 0 } {
	set fileid stdout
    } else {
	if [catch {open $outfile w} fileid] {
	    puts stderr \
		    "Unable to open machine header file $outfile for writing"
	    return 0
	}
    }

    # Dump header info
    puts $fileid \
{/* FILE: port.h               -*-Mode: c++-*-
 *
 * Machine specific #define's and typedef's, generated by [Oc_MakePortHeader]
 *
 */

#ifndef _OC_PORT
#define _OC_PORT

#include <math.h>
#include <float.h>
#define MEMBLOCKSIZE 4096 /* Natural system memory blocksize.  Shouldn't  */
               /* affect program correctness, but may affect performance. */

/* End includes */}

    foreach {tclmajor tclminor} [split [info tclversion] .] {}
    puts $fileid "
#define CONFIG_TCL_MAJOR_VERSION $tclmajor
#define CONFIG_TCL_MINOR_VERSION $tclminor
#define CONFIG_TCL_PATCH_LEVEL \"[info patchlevel]\"
"

    # Write universal (I hope!) typedef's
    puts $fileid "/* Variable type declarations.  The '****#m' */"
    puts $fileid "/* types are at *least* '#' bytes wide.      */"
    puts $fileid "typedef  int                 BOOL;"
    puts $fileid "typedef  unsigned char       BYTE;"
    puts $fileid "typedef  char                CHAR;"
    puts $fileid "/* typedef  signed char        SCHAR; */"
    puts $fileid "/** signed type not supported by HP C++ compiler **/"
    puts $fileid "typedef  unsigned char      UCHAR;"

    # Write float typedef's
    if { $varsize(float) != $varsize(double) } {
	puts $fileid "typedef  float              REAL$varsize(float);"
    }
    puts $fileid "typedef  double             REAL$varsize(double);"
    if { $varsize(float) >= 4 } {
	puts $fileid "typedef  float              REAL4m;"
    }
    if { $varsize(float) >= 8 } {
	puts $fileid "typedef  float              REAL8m;"
    } elseif { $varsize(double) >= 8 } {
	puts $fileid "typedef  double             REAL8m;"
    }
    puts $fileid "#define REAL4_EPSILON       $vareps(FLT)"
    puts $fileid "#define SQRT_REAL4_EPSILON  $vareps(SQRT_FLT)"
    puts $fileid "#define REAL8_EPSILON       $vareps(DBL)"
    puts $fileid "#define SQRT_REAL8_EPSILON  $vareps(SQRT_DBL)"
    # May add "long double" support in the future

    # Write integer typedef's
    if { $varsize(short) < $varsize(int) } {
	puts $fileid "typedef  short               INT$varsize(short);"
	puts $fileid "typedef  unsigned short     UINT$varsize(short);"
    }
    puts $fileid "typedef  int                 INT$varsize(int);"
    puts $fileid "typedef  unsigned int       UINT$varsize(int);"
    if { $varsize(long) > $varsize(int) } {
	puts $fileid "typedef  long                INT$varsize(long);"
	puts $fileid "typedef  unsigned long      UINT$varsize(long);"
    }
    foreach msize { 2 4 8 16 } {
	if { $varsize(int) >= $msize } {
	    # Use type "int" if possible, as this is likely
	    # to be the preferred machine word size
	    puts $fileid "typedef  int                 INT${msize}m;"
	    puts $fileid "typedef  unsigned int       UINT${msize}m;"
	} elseif { $varsize(long) >=$msize } {
	    # Otherwise, fall back on long type
	    puts $fileid "typedef  long                INT${msize}m;"
	    puts $fileid "typedef  unsigned long      UINT${msize}m;"
	}
	# In the future, may add support for "long long" integral type
    }

    # Pointers
    puts $fileid "\n#define POINTERWIDTH $varsize(pointer)"

    # Byte order.  For now just use 4-byte wide ordering
    foreach vartype { int long short float double } {
	if { $varsize($vartype) == 4 } {
	    puts $fileid "#define BYTEORDER $varorder($vartype)"
	    break
	}
    }

    # Machine platform types
    puts $fileid "\n/* System type info */"
    puts $fileid "#define UNIX 1"
    puts $fileid "#define WINDOWS 2"
    puts $fileid "#define MACINTOSH 3"
    # Note: Local variable "systemtype" is set at top of this proc
    if {[string compare $systemtype "unix"] == 0} {
	puts $fileid "#define SYSTEM_TYPE UNIX"
    } elseif {[string compare $systemtype "windows"] == 0} {
	puts $fileid "#define SYSTEM_TYPE WINDOWS"
    } elseif {[string compare $systemtype "macintosh"] == 0} {
	puts $fileid "#define SYSTEM_TYPE MACINTOSH"
    }

    set OS [string tolower [string trim $tcl_platform(os)]]
    set OSVERSION [string tolower [string trim $tcl_platform(osVersion)]]
    set OSMAJOR $OSVERSION
    regexp {([^.]*).*} $OSVERSION match OSMAJOR
# Random number generator protoypes
# 	Should later replace this with an Autoconf-style determination.
    if {[string match sunos $OS] && $OSMAJOR >= 5} {
        puts $fileid {
/* Random number generator.  Replace with your own if you don't like    */
/* random().  Prototypes for random() and srandom() appear to be        */
/* missing from the Solaris system header files (though they are in the */
/* runtime library).  If you don't have random(), you can use rand()    */
/* instead, but be wary of some older, badly broken versions of rand(). */
extern "C" {
   /* If you get error messages about multiple inconsistent definitions  */
   /* of random or srandom, just yank out this prototype block.          */
   long random(void);
   void srandom(unsigned int seed);
}
#define OMF_SRANDOM(seed)  srandom(seed)
#define OMF_RANDOM()       random()
#define OMF_RANDOM_MAX     0x7FFFFFFF /* Is this system dependent? */
}
    } elseif {[string match sunos $OS] && $OSMAJOR < 5} {
        puts $fileid {
/* Random number generator --special case for sunbow (SunOS 4.1.3). */
#define OMF_SRANDOM(seed)  srand(seed)
#define OMF_RANDOM()       rand()
#define OMF_RANDOM_MAX     0x7FFF
}
    } elseif {[string match windows $systemtype]} {
        puts $fileid {
/* Random number generator. Replace with your own if */
/* if you don't like rand().                        */
#define OMF_SRANDOM(seed)  srand(seed)
#define OMF_RANDOM()       rand()
#define OMF_RANDOM_MAX     RAND_MAX
}
    } elseif {[string match aix $OS]} {
        puts $fileid {
/* Random number generator. Replace with your own if */
/* if you don't like rand().                        */
#define OMF_SRANDOM(seed)  srand(seed)
#define OMF_RANDOM()       rand()
#define OMF_RANDOM_MAX     RAND_MAX
}
    } elseif {[string match hp-ux $OS]} {
        puts $fileid {
/* Random number generator. Replace with your own if */
/* if you don't like rand().                        */
#define OMF_SRANDOM(seed)  srand(seed)
#define OMF_RANDOM()       rand()
#define OMF_RANDOM_MAX     RAND_MAX
}
    } else {
        puts $fileid {
/* Random number generator.  Replace with your own if */
/* you don't like random().                           */ 
#define OMF_SRANDOM(seed)  srandom(seed)
#define OMF_RANDOM()       random()
#define OMF_RANDOM_MAX     0x7FFFFFFF /* Is this system dependent? */
}
    }

    # Windows vs. Unix-isms
    if {[string compare $systemtype unix] == 0} {
        puts $fileid {
/* For unix */
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>  /* Child process cleanup */

/* NICE_DEFAULT is the value passed to nice() inside  */
/* MakeNice().                                        */
#define NICE_DEFAULT 9
/* If your system doesn't have nice, uncomment the next line, or put */
/* in a suitable replacement macro (using, perhaps, setpriority()?). */
/*#define nice(x) 0 */

/* Directory path separator; Unix uses '/', DOS uses '\'. */
/* Rumor has it that the Mac uses ':'.                    */
#define DIRSEPCHAR '/'
#define DIRSEPSTR  "/"
#define PATHSPLITSTR ":"
}
    }
    if {[string compare $systemtype windows] == 0} {
        puts $fileid {
/* For Windows */
#define rint(x) ((x)>=0 ? double(int(x+0.5)) : double(int(x-0.5)))
 
/* NICE_DEFAULT is the priority level passed to SetPriorityClass() */
/* inside MakeNice().                                              */
#define NICE_DEFAULT IDLE_PRIORITY_CLASS
 
/* Windows header file.  NB: This defines a lot of stuff we      */
/* don't need or really want, like macros min(x,y) and max(x,y). */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN
 
/* Directory path separator; Unix uses '/', DOS uses '\'. */
/* Rumor has it that the Mac uses ':'.                    */
#define DIRSEPCHAR '\\'
#define DIRSEPSTR  "\\"
#define PATHSPLITSTR ";"
}
    }

   # Dump trailer
   puts $fileid "#endif /* _OC_PORT_H */"

    if { [string compare $outfile stdout] != 0 } {
	close $fileid
    }
}

# Tk bindings are susceptible to reentrancy problems.  For example,
# suppose a user double clicks on a single click binding.  Two
# single clicks events get entered into the event loop.  Now, if
# during the processing of the binding on the first click, the
# event loop were to be re-entered, then this would cause a second
# call to the binding (from the second click) before the first call
# has finished processing.  In particular, one should be aware that
# anytime the binding proc accesses a global variable, that global
# variable may have a trace on it, and that trace may make an update,
# tkwait, or similar request...against the better interests of the
# binding proc.
#   To protect against reentrancy, use the following proc,
# OMF_ThreadSafe.  You will probably want to use a unique "thread_id"
# for each protected code segment.  Ideally, "script" should just be
# a proc call, in which case the name of the proc is a good choice for
# the thread_id.  (Note: The "wait" retry time is in milliseconds.)
#   Sample usage:
#
#    button $w.wFDokbtn -text "OK" \
#           -command "OMF_ThreadSafe OMD_FDokcmd,$w \{OMF_FDokcmd $w\}"
#
# Here I want to protect against reentrancy from the same window, $w,
# so I append that as an additional identifier on thread_id.
#
# NOTE 1: The script is executed at global scope
# NOTE 2: In the current implementation, order of processing of delayed
#         calls may not be preserved.
proc Oc_ThreadSafe { thread_id script {wait 500}} {
    global omfThreadLock errorInfo errorCode
    if { [info exists omfThreadLock($thread_id)] \
            && $omfThreadLock($thread_id)==1 } {
        # Thread locked.  Put script on shelf and try again later
        after $wait [list Oc_ThreadSafe $thread_id $script]
    } else {
        # Otherwise, process the script now
        set omfThreadLock($thread_id) 1
        set errcode [catch { uplevel #0 $script } errmsg]
        set omfThreadLock($thread_id) 0
        if { $errcode != 0 } {
            error $errmsg $errorInfo $errorCode
        }
    }    
}

# Some event bindings, in particular mouse drag events, can generate a
# nearly continuous stream of events in response to user input.  If the
# handler for these events is slow, then these events can pile up in the
# event loop.  It is often the case in this situation that all
# intermediate events can be ignored, and only the last one processed.
# This proc takes a standard bind command, and puts a wrapper around it
# to implement this behavior.  The returned string can be directly bound
# to an event.  NOTE: This routine does not protect against re-entrancy
# on $cmd.  If $cmd services the event loop, then there is the
# possibility of event re-ordering.  In this circumstance, use
# Oc_SafeSkipWrap instead.
set _oc_skipwrap(count) 0
proc Oc_SkipWrap { cmd {wait 3}} {
    global _oc_skipwrap   ;# Wrapper state
    set id $_oc_skipwrap(count) ;# Elt to hold pending event id
    incr _oc_skipwrap(count)
    set _oc_skipwrap($id) {} ;# Initialize
    set newcmd [format {
        global _oc_skipwrap
        after cancel $_oc_skipwrap(%s)
        set _oc_skipwrap(%s) [after %s {%s}]
    } $id $id $wait $cmd]  ;# Create wrapped command
    return $newcmd
}

# Oc_SafeSkipWrap is a version of Oc_SkipWrap that puts a
# semaphore-style lock around the wrapped command.  Before a command is
# executed, a check is made to see if a lock is set.  If so, the command
# goes back onto the event queue.  If not, the lock is set, the command
# is run, and then the lock is unset.  Three points: 1) It is important
# that the command run to completion.  If it should fail, then the lock
# would never be reset and all future commands on this binding would be
# locked out. Because of this, it is not a bad idea to wrap $cmd up
# inside a 'catch'.  2) The locking mechanism protects $cmd from
# reentrancy from _this_ binding.  For example, say $cmd is a call to
# proc foo.  This lock is placed outside of foo, so even if the lock is
# set as a result of this binding, that does not lock out calls to foo
# from a different event binding.  If you need absolute reentrancy
# protection, use Oc_ThreadSafe. 3) Because only 1 event is queued up at
# a time (unprocessed events are thrown away when a new one is
# generated), event order _is_ preserved on execution.
set _oc_safeskipwrap(count) 0
proc Oc_SafeSkipWrap { cmd {wait 3}} {
    global _oc_safeskipwrap   ;# Wrapper state
    set id $_oc_safeskipwrap(count) ;# Elt to hold pending event id
    incr _oc_safeskipwrap(count)
    set _oc_safeskipwrap($id) {} ;# Initialize
    set _oc_safeskipwrap(lock$id) {}
    set newcmd [format {
        global _oc_safeskipwrap
        after cancel $_oc_safeskipwrap(%s)
        set _oc_safeskipwrap(%s) \
                [after %s {Oc_SafeSkipLock %s {%s} %s}]
    } $id $id $wait $id $cmd $wait]  ;# Create wrapped command
    return $newcmd
}
proc Oc_SafeSkipLock { id cmd wait } {
    # Processes ripened events
    global _oc_safeskipwrap
    switch {} $_oc_safeskipwrap(lock$id) {
        set _oc_safeskipwrap(lock$id) 1
        eval $cmd
        set _oc_safeskipwrap(lock$id) {}
    } default {
        set _oc_safeskipwrap($id) \
                [after $wait "Oc_SafeSkipLock $id \{$cmd\} $wait"]
    }
}


proc Oc_TempName { {basedir {}} {baseprefix {_}} {suffix {}}} {
    if  {[string match {} $basedir]} {
        Oc_TempFile New f -stem $baseprefix -extension $suffix
    } else {
        Oc_TempFile New f -stem $baseprefix -extension $suffix \
                -directory $basedir
    }
    set retval [$f AbsoluteName]
    $f Delete
    return $retval
}

proc Oc_IgnoreInteractiveSignals {} {
    catch {Oc_IgnoreSignal  2}  ;# Ctrl-C generates SIGINT
    catch {Oc_IgnoreSignal  3}  ;# Ctrl-\ generates SIGQUIT
    catch {Oc_IgnoreSignal 18}  ;# Ctrl-Z generates SIGTSTP
    catch {Oc_IgnoreSignal 20}  ;# Invoke automatic child reaping.  This
                                ## may not work on all platforms.
    # Unix list:
    #  1) SIGHUP       2) SIGINT       3) SIGQUIT      4) SIGILL
    #  5) SIGTRAP      6) SIGIOT       7) SIGEMT       8) SIGFPE
    #  9) SIGKILL     10) SIGBUS      11) SIGSEGV     12) SIGSYS
    # 13) SIGPIPE     14) SIGALRM     15) SIGTERM     16) SIGURG
    # 17) SIGSTOP     18) SIGTSTP     19) SIGCONT     20) SIGCHLD
    # 21) SIGTTIN     22) SIGTTOU     23) SIGIO       24) SIGXCPU
    # 25) SIGXFSZ     26) SIGVTALRM   27) SIGPROF     28) SIGWINCH
    # 29) SIGPWR      30) SIGUSR1     31) SIGUSR2
    # Under Windows, only the following subset is apparently available:
    #   SIGINT, SIGILL, SIGFPE, SIGSEGV, SIGTERM, SIGBREAK, SIGABRT
}

proc Oc_StackTrace {} {
    set history {}
    for {set n [expr [info level]-1]} {$n>0} {incr n -1} {
        append history "LEVEL $n: [info level $n]\n\n"
    }
    return $history
}

